#include <windows.h>
#include <tchar.h>

#define DATA_AVAILABLE	 (0)
#define DATA_UNAVAILABLE (1)
#define DATA_COMPLETE    (2)

#define MBOX_SIZE	(1024)

TCHAR* inputFile = _T("bar.txt");

// pojmenovani sdilenych objektu
TCHAR* mboxName = _T("mbox");
TCHAR* mboxLockName = _T("mboxlock");

HANDLE hLock;           // sdileny semafor
HANDLE hFile;
HANDLE hMapping;

/** struktura reprezentujici schranku, do ktere je mozne ulozit nanejvys jednu zpravu */
struct mbox {
	char status;          // stav schranky (viz makra DATA_*)
	char data[MBOX_SIZE]; // samotna data
};

/** inicializuje obsah schranky */
struct mbox* mbox_init()
{
	hLock = CreateMutex(NULL, FALSE, mboxLockName);

	hFile = CreateFile(inputFile, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		_tprintf(_T("Unable to open: %s\n"), inputFile);
		return NULL;
	}

	hMapping = CreateFileMapping(hFile, NULL, PAGE_READWRITE, 0, sizeof(struct mbox), mboxName);
	if (hMapping == NULL) {
		_tprintf(_T("Unable to create file mapping\n"));
		_tprintf(_T("last error: %i"), GetLastError());
		return NULL;
	}
	

	/*
	// vyuziti anonymni pameti
	hMapping = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, sizeof(struct mbox), mboxName);
	if (hMapping == NULL) {
		_tprintf(_T("Unable to create file mapping\n"));
		_tprintf(_T("last error: %i"), GetLastError());
		return NULL;
	}
	*/
	struct mbox* mbox = MapViewOfFile(hMapping, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(struct mbox));
	mbox->status = DATA_UNAVAILABLE;

	return mbox;
}

/** otevre existujici schranku */
struct mbox* mbox_open()
{
	hLock = OpenMutex(SYNCHRONIZE, FALSE, mboxLockName);

	if (!hLock) {
		_tprintf(_T("Unable to open lock: %s\n"), mboxLockName);
		return NULL;
	}

	hMapping = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, mboxName);
	if (hMapping == NULL) {
		_tprintf(_T("Unable to open file mapping\n"));
		_tprintf(_T("last error: %i"), GetLastError());
		return NULL;
	}

	struct mbox* mbox = MapViewOfFile(hMapping, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(struct mbox));

	return mbox;
}

/**
 * Pokud je schranka prazdna, ulozi do ni zpravu v podobe celeho sicla.
 * V opacnem pripade ceka na uvolneni schranky.
 */
void mbox_send(struct mbox* mbox, int value)
{
	int send = 0;
	while (!send) {
		WaitForSingleObject(hLock, INFINITE);
		if (mbox->status == DATA_UNAVAILABLE) {
			_stprintf_s(mbox->data, MBOX_SIZE, _T("msg: %i"), value);
			mbox->status = DATA_AVAILABLE;
			_tprintf(_T("SENDER: %s\n"), mbox->data);
			send = 1;
		}
		ReleaseMutex(hLock);
//		Sleep(1000);
	}
}

/**
 * Ze schranky vyzvedna zpravu a vypise ji na std. vystup.
 * Pokud zprava ve schrance neni, ceka.
 * Navratovou hodnotou je signalizovano uspesne precteni dat,
 * nebo konec komunikace.
 */
int mbox_receive(struct mbox* mbox)
{
	int read = 0; // signalizuje uspesne precteni dat
	int prev_status = 0;
	while (!read) {
		WaitForSingleObject(hLock, INFINITE);
		prev_status = mbox->status;
		switch (mbox->status) {
		case DATA_AVAILABLE:
			_tprintf(_T("%s\n"), mbox->data);
		case DATA_COMPLETE:
			mbox->status = DATA_UNAVAILABLE;
			read = 1;
			break;
		}
		ReleaseMutex(hLock);
	}
	return prev_status;
}

void mbox_destroy(struct mbox* mbox)
{
	int success = 0;
	// signalizuje konec komunikace
	while (!success) {
		WaitForSingleObject(hLock, INFINITE);
		if (mbox->status == DATA_UNAVAILABLE) {
			mbox->status = DATA_COMPLETE;
			success = 1;
		}
		ReleaseMutex(hLock);
	}

	// ceka, az prijemce zpravu o konci komunikace obdrzi
	while (!success) {
		WaitForSingleObject(hLock, INFINITE);
		if (mbox->status == DATA_UNAVAILABLE) {
			success = 1;
		}
		ReleaseMutex(hLock);
	}

	CloseHandle(hFile);
}

static void show_help(TCHAR* app_name)
{
	_tprintf(_T("Usage: %s [writer|reader]\n"), app_name);
}

int _tmain(int argc, TCHAR** argv)
{
	if (argc < 2) {
		show_help(argv[0]);
		return 0;
	}

	struct mbox* mb;

	if (!_tcscmp(argv[1], _T("writer"))) {

		// inicializuje mbox
		mb = mbox_init();
		_tprintf(_T("mbox at:%p\n"), mb);

		// zasila cisla z interval 0 ... 99
		for (int i = 0; i < 100; i++) {
			mbox_send(mb, i);
			//sleep(1);
		}
		// odstrani mbox
		mbox_destroy(mb);
		UnmapViewOfFile(mb);
		CloseHandle(hFile);
		CloseHandle(hLock);
		CloseHandle(hMapping);
	}
	else if (!_tcscmp(argv[1], _T("reader"))) {
		mb = mbox_open();
		_tprintf(_T("mbox at:%p\n"), mb);
		int status;
		do {
			// prijima zpravy
			status = mbox_receive(mb);
		} while (status != DATA_COMPLETE);
		UnmapViewOfFile(mb);
		CloseHandle(hLock);
		CloseHandle(hMapping);
	}
	else {
		show_help(argv[0]);
	}

	return 0;

}

